/**
 * Copyright (C) 2009 DATAPREV-ES
 * @author Vinicius Avellar Moreira
 * Classe singleton responsvel pela autenticao no gerente web.
 */

#ifndef _CACIC_AUTH_
#define _CACIC_AUTH_

#include <vector>
#include <string>
using namespace std;

#include "supInfoDlg.h"

class CACIC_Auth {

public:

	/**
	*Autoriza reconexao do suporte remoto.
	*/
	static bool autorizaReconexao(char nm_usuario_cli[],char te_senha_cli[]);

	/**
	*Deleta arquivo aps 5 minutos da queda do servio.
	*/
	void deletaDataServer();

	/**
	*Verifica consistncia do arquivo - token
	**/
	static BOOL verificaArq(string name, string value, FILE* arq);
	static BOOL verificaArq(string name, FILE* arq);


	/**
	 * Nome do arquivo temporrio de trava.
	 * Utilizado para que o cacic impea a execuo de coletas com o suporte ativo.
	 */
	static const string AGUARDE_FILENAME;

	/**
	 * Tempo mximo que o servidor pode ficar aberto sem conexes. (mins)
	 */
	static const UINT TEMPO_IDLE;

	/** Lista de usurios cliente. */
	vector<ClienteSRC> m_listaClientes;

	/** ltimo usurio cliente conectado. */
	ClienteSRC m_novoCliente;

	/**
	 * Janela com informaes sobre o cliente conectado.
	 * Exibida enquanto h suporte em andamento.
	 */
	supInfoDlg m_infoDlg;
 
	/**
	 * Varivel de deciso sobre o logout do sistema aps o suporte.
	 */
	bool m_efetuarLogout;

	/** Singleton. */
	static CACIC_Auth* getInstance()
	{
		if (!m_instance) m_instance = new CACIC_Auth();
		return m_instance;
	}

	/* MTODOS DE ENCAPSULAMENTO --> */
	void setServidorWeb(string newServidorWeb) {m_servidorWeb = newServidorWeb;}
	string getServidorWeb() {return m_servidorWeb;}
	void setScriptsPath(string newScriptsPath) {m_scriptsPath = newScriptsPath;}
	void setTempPath(string newTempPath) {m_tempPath = newTempPath;}
	void setSOVersion(char* newSOVersion) {m_soVersion = newSOVersion;}
	void setNodeAdress(char* newNodeAdress) {m_nodeAdress = newNodeAdress;}
	void setPalavraChave(const char* newPalavraChave) {m_palavraChave = newPalavraChave;}
	void setPorta(UINT newPorta) {m_porta = newPorta;}
	UINT getPorta() {return m_porta;}
	void setTimeout(UINT newNuTimeout) {m_nuTimeout = newNuTimeout;}
	UINT getTimeout() {return m_nuTimeout;}
	// static void setFilePathGlobal(string newPath) {FILE_PATH_GLOBAL = newPath;}
	// static string getFilePathGlobal() {return FILE_PATH_GLOBAL;}
	
	/* <-- MTODOS DE ENCAPSULAMENTO */

	/**
	 * Retorna os valores padro de post, usados na 
	 * comunicao com o gerente web.
	 * te_so, te_node_address, te_palavra_chave
	 * @return String com o post padro formatado.
	 */
	string getPostComum();

	/**
	 * Remove o usurio cliente da lista.
	 * @param vncCID ID do cliente VNC, utilizado para
	 * diferenciar os clientes, caso haja mais de um.
	 */
	void removeCliente(short vncCID);

	/**
	 * Faz a comunicao com o gerente web para validar a palavra chave
	 * e criar uma nova sesso para o suporte remoto.
	 * @return bool Status da autenticao.
	 */
	bool autentica();

	/**
	 * Se comunica com o gerente web para validar o usurio cliente.
	 * Se o usurio for vlido, ele cria uma nova sesso de conexo.
	 * @param nm_usuario_cli String codificada contendo o nome de usurio.
	 * @param te_senha_cli String codificada contendo a senha do usurio.
	 * @param te_node_address_cli String codificada contendo o MAC address do cliente.
	 * @param te_documento_referencial String codificada contendo o Documento de Referncia do suporte remoto.
	 * @param te_motivo_conexao String codificada contendo o motivo do suporte remoto.
	 * @param te_so_cli String codificada contendo a identificao do SO do cliente.
	 * @param vncCID ID do cliente VNC.
	 * @param peerName String contendo o endereo ip do cliente.
	 */
	bool validaTecnico(char nm_usuario_cli[], char te_senha_cli[], char te_node_address_cli[],
					   char te_documento_referencial[], char te_motivo_conexao[], char te_so_cli[], 
					   const short vncCID, const char peerName[]);
	
	/**
	 * Se comunica com o gerente web para atualizar a sesso de suporte.
	 */
	void atualizaSessao();

	/**
	 * Verifica se a ltima sesso foi finalizada h mais de 30 minutos e, caso positivo
	 * remove o arquivo ck_conexao.ini
	 */
	void CACIC_Auth::verifyTimeOutCon();

	/**
	 * Envia o log do chat para o gerente web durante o suporte remoto.
	 * @param te_mensagem Mensagem recebida/enviada.
	 * @param cs_origem Origem da mensagem, cliente/servidor.
	 */
	void sendChatText(char te_mensagem[], char cs_origem[]); 
	
	/** Fecha o servidor. */
	void finalizaServidor();

	string getTempPath(){return m_tempPath;}


private:

	// static string FILE_PATH_GLOBAL;

	/** Singleton. */
	static CACIC_Auth* m_instance;

	CACIC_Auth() {
		m_idleTime = TEMPO_IDLE;
		m_efetuarLogout = true;
	}

	virtual ~CACIC_Auth() {}

	/** Endereo do servidor web. */
	string m_servidorWeb;
	/** Caminho dos scripts no servidor web. */
	string m_scriptsPath;
	/** Caminho esttico para a pasta temp do cacic. */
	string m_tempPath;

	/** Usurio host do suporte. */
	string m_usuario;
	/** ID da sesso iniciada pelo usurio host. */
	string m_idSessao;

	/** Verso do sistema operacional do host. */
	string m_soVersion;
	/** MAC Address do host. */
	string m_nodeAdress;
	/** Palavra chave. Utilizada na comunicao com o gerente web. */
	string m_palavraChave;

	/** Porta de escuta. */
	UINT m_porta;
	/** Tempo limite que o srcacic pode ficar ocioso antes de fechar-se. */
	UINT m_nuTimeout;
	/** Tempo que o servidor est ocioso */
	UINT m_idleTime;

	/** Nome do script de configuraes do gerente web. */
	static const string GET_CONFIG_SCRIPT;
	/** Nome do script de sesses do gerente web. */
	static const string SET_SESSION_SCRIPT;
	/** Nome do script de autenticao do gerente web. */
	static const string AUTH_CLIENT_SCRIPT;
	/** Tamanho padro da resposta recebida pela requisio http. */
	static const unsigned int TAMANHO_RESPOSTA;
	/** Nome do arquivo temporrio de atualizao da palavra chave. */
	static const string COOKIE_FILENAME;
	/** Nome do arquivo temporario para reconexo do suporte remoto. */
	static const string TOKEN_FILENAME;

	/**
	 * Apresenta o dilogo de login do usurio host e
	 * valida os dados no gerente web.
	 * @param listaDominios Lista de domnios obtida na autenticao.
	 */
	bool autenticaUsuario(vector<Dominio> &listaDominios);

	/**
	 * Verifica a autenticao da chave no gerente web.
	 * @param resposta Resposta XML gerada na comunicao com o gerente web.
	 * @param listaDominios Lista de domnios obtida na autenticao.
	 */
	bool verificaAuthChave(char resposta[], vector<Dominio> &listaDominios);

	/**
	 * Verifica se a resposta da autenticao do usurio host foi positiva.
	 * @param resposta Resposta XML gerada na comunicao com o gerente web.
	 */
	bool verificaAuthDominio(char resposta[]);

	/**
	 * Verifica se a resposta da autenticao do tcnico foi positiva,
	 * armazena o novo cliente na lista e exibe a tela de informaes do suporte.
	 * @param nm_usuario_cli String codificada contendo o nome de usurio.
	 * @param te_senha_cli String codificada contendo a senha do usurio.
	 * @param te_node_address_cli String codificada contendo o MAC address do cliente.
	 * @param te_documento_referencial String codificada contendo o Documento de Referncia do suporte remoto.
	 * @param te_motivo_conexao String codificada contendo o motivo do suporte remoto.
	 * @param te_so_cli String codificada contendo a identificao do SO do cliente.
	 * @param vncCID ID do cliente VNC.
	 * @param peerName String contendo o endereo ip do cliente.
	 */
	bool verificaAuthTecnico(char resposta[], char te_node_address_cli[], char te_documento_referencial[],
							 char te_motivo_conexao[], char te_so_cli[], 
							 const short vncCID, const char peerName[]);

	/**
	 * Verifica o valor de retorno STATUS que  enviado pelo gerente web
	 * aps cada comunicao para confirmar a operao.
	 * <b>Valores retornados:</b><br />
	 * OK: A operao teve xito.<br />ERRO: A operao falhou.
	 * @param resposta Resposta XML gerada na comunicao com o gerente web.
	 */
	bool verificaStatus(char resposta[]);
};

#endif
