/**
 * Copyright (C) 2009 DATAPREV-ES
 * @author Vinicius Avellar Moreira
 * Classe com alguns mtodos utilitrios.
 */

#ifndef _CACIC_UTILS_
#define _CACIC_UTILS_

#include <string>
using namespace std;

#include <windows.h>

/**
 * Struct referente a um usurio cliente.
 */
struct ClienteSRC {
	short vncCID;
	string peerName;
	
	string id_usuario_visitante; 
	string id_usuario_cli; 
	string id_conexao; 
	string nm_usuario_completo; 
	string te_node_address_visitante; 
	string te_node_address_cli; 
	string te_documento_referencial; 
	string te_motivo_conexao; 
	string te_so_visitante; 
	string te_so_cli; 
	string dt_hr_inicio_sessao; 
};

/**
 * Struct referente a um domnio de autenticao.
 */
struct Dominio {
	Dominio(string p_id, string p_nome) : id(p_id), nome(p_nome) {}
	Dominio() : id(""), nome("") {}
	Dominio(const Dominio& d) : id(d.id), nome(d.nome) {}
	string id;
	string nome;
};

/**
* Funo de Comparao de objetos de mesmo tipo. Resultados possveis:
* maior, menor ou igual.
*/
template <typename T>
int Compare (T a, T b){
	if (a > b) return 1;
	else if (a < b) return -1;
	return 0;
};

	/**
	* Funo que compara Datas do sistema, atraves da funo declarada
	* anteriormente: Compare.
	*/
	//int DateCompare (SYSTEMTIME d1, SYSTEMTIME d2);

class CACIC_Utils {

public:

	/** Fonte padro usado nos dilogos. */
	static const string F_SANS_SERIF;

	/**
	 * Mtodo bruto para ler uma tag especfica de um arquivo xml.
	 * @param xml String no formato de arquivo xml.
	 * @param tagname String com o nome da tag a ser pesquisada.
	 * @param conteudo String com o contedo da tag pesquisada.
	 * @trows CACIC_Exception caso a tag no seja encontrada.
	 */
	static void leTag(char xml[], char tagname[], string &conteudo);

	/**
	 * Troca caracteres especficos de uma string.
	 * @param str String a ser modificada.
	 * @param key String com o caractere ou conjunto de caracteres que sero substitudos.
	 * @param newKey String com o caractere ou conjunto de caracteres que iro substituir.
	 */
	static void replaceAll(string &str, string key, string newkey);

	/**
	 * Codifica a string, removendo os caracteres especiais por %cdigo dos mesmos.
	 * @param decoded String que ser codificada.
	 */
	static void urlEncode(string &decoded);

	/**
	 * Decodifica a string, retornando os cdigos dos caracteres pelos prprios caracteres.
	 * @param encoded String que ser decodificada.
	 */
	static void urlDecode(string &encoded);

	/**
	 * Mesma funo do urlEncode, porm os caracteres sero substitudos
	 * por tags especficas, e no pelo cdigo.
	 * @param entrada String que ser codificada.
	 */
	static void simpleUrlEncode(string &entrada);

	/**
	 * Faz o inverso do simpleUrlEncode, trocando as tags especficas pelos
	 * respectivos caracteres.
	 * @param entrada String que ser codificada.
	 */
	static void simpleUrlDecode(string &entrada);

	/**
	 * Transforma o byte em codigo ascii, retornando o char correspondente.
	 * @param first Primeiro hexa do caractere.
	 * @param second Segundo hexa do caractere.
	 * @return Char correspondente ao cdigo ascci encontrado.
	 */
	static char hexToAscii(char first, char second);

	/**
	 * Retira os espaos em branco do comeo e fim da string.
	 * @param str String a ser modificada.
	 */
	static void trim(string &str);

	/**
	 * Mtodo para alterar a fonte de um determinado elemento de um dilogo.
	 * @param dlgHandle Handler do dilogo.
	 * @param dlgItem Item do dilogo que ter a fonte trocada.
	 * @param fontSize Tamanho da fonte.
	 * @param fontName Nome da fonte.
	 * @param fontIsBold Define o peso da fonte: true = bold, false = normal.
	 */
	static void changeFont(HWND dlgHandle, int dlgItem, 
						   int fontSize, string fontName, 
						   bool fontIsBold = false);

	/**
	 * Mtodo para separar a string em partes delimitadas por um, ou um conjunto,
	 * de caracteres.
	 * @param str String a ser tokenizada.
	 * @param tokens Vetor de sada dos tokens gerados.
	 * @param delimiters Delimitadores que sero usados para separar a string.
	 * @note http://www.linuxselfhelp.com/HOWTO/C++Programming-HOWTO-7.html
	 */
	//static void tokenize(const string &str, vector<string> &tokens, const string &delimiters = " ");

	/**
	 * Obtm o MAC Address da placa de rede.<br />
	 * TODO: Quando houver mais de uma placa de rede no pc, verificar qual
	 * est se comunicando com o servidor para enviar o MAC certo.
	 */
	static string getMACAddress();

	/**
	 * Obtm a identificao do sistema operacional.<br />
	 * Artigo sobre SOID:<br />
	 * http://www.codeguru.com/cpp/w-p/system/systeminformation/article.php/c8973__2/
	 */
	static string getSOID();

	/**
	* Funo que compara Datas do sistema, atraves da funo declarada
	* anteriormente: Compare.
	*/
	static int DateCompare (SYSTEMTIME d1, SYSTEMTIME d2);

private:

	/**	
	 * Este mtodo virtual puro  um truque para que a classe
	 * se torne abstrata e no possa ser instanciada.
	 */
	virtual void cutils() = 0;

};

#endif